"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const core_1 = require("@stylable/core");
const vscode_uri_1 = require("vscode-uri");
const provider_1 = require("./provider");
const completion_providers_1 = require("./completion-providers");
const diagnosis_1 = require("./diagnosis");
const provider_2 = require("./provider");
const css_service_1 = require("../model/css-service");
const color_provider_1 = require("./feature/color-provider");
const dedupe_refs_1 = require("./dedupe-refs");
const typescript_support_1 = require("./typescript-support");
function wrapFs(fs, docs) {
    const readFileSync = ((path, ...args) => {
        const file = docs.get(vscode_uri_1.URI.file(path).toString());
        return file ? file.getText() : fs.readFileSync(path, ...args);
    });
    return { ...fs, readFileSync };
}
class StylableLanguageService {
    constructor({ rootPath, fs, requireModule, textDocuments }) {
        this.docsDispatcher = textDocuments;
        this.fs = wrapFs(fs, this.docsDispatcher);
        this.tsLanguageService = typescript_support_1.typescriptSupport(this.fs);
        this.stylable = new core_1.Stylable(rootPath, this.fs, requireModule);
        this.provider = new provider_1.Provider(this.stylable, this.tsLanguageService);
        this.cssService = new css_service_1.CssService(this.fs);
    }
    getDocsDispatcher() {
        return this.docsDispatcher;
    }
    getStylable() {
        return this.stylable;
    }
    getFs() {
        return this.fs;
    }
    provideCompletionItemsFromSrc(src, pos, fileName) {
        return this.provider.provideCompletionItemsFromSrc(src, pos, fileName, this.fs);
    }
    async getDefinitionLocation(src, position, filePath) {
        const defs = await this.provider.getDefinitionLocation(src, position, vscode_uri_1.URI.file(filePath).fsPath, this.fs);
        return defs.map(loc => vscode_languageserver_types_1.Location.create(vscode_uri_1.URI.file(loc.uri).fsPath, loc.range));
    }
    getSignatureHelp(src, pos, filePath, paramInfo) {
        return this.provider.getSignatureHelp(src, pos, filePath, this.fs, paramInfo);
    }
    getRefs(filePath, position) {
        return provider_2.getRefs(filePath, position, this.fs, this.stylable);
    }
    resolveDocumentColors(document) {
        return color_provider_1.resolveDocumentColors(this.stylable, this.cssService, document, this.fs);
    }
    getColorPresentation(document, params) {
        return color_provider_1.getColorPresentation(this.cssService, document, params);
    }
    diagnose(connection) {
        const diagnoseConfig = {
            connection,
            cssService: this.cssService,
            docsDispatcher: this.docsDispatcher,
            stylable: this.stylable
        };
        return () => diagnose(diagnoseConfig);
    }
    onCompletion(params) {
        const documentUri = params.textDocument.uri;
        const position = params.position;
        if (!documentUri.endsWith('.st.css') && !documentUri.startsWith('untitled:')) {
            return [];
        }
        const document = this.docsDispatcher.get(documentUri);
        if (document) {
            const res = this.provider.provideCompletionItemsFromSrc(document.getText(), {
                line: position.line,
                character: position.character
            }, vscode_uri_1.URI.parse(documentUri).fsPath, this.fs);
            return res
                .map((com) => {
                const lspCompletion = vscode_languageserver_types_1.CompletionItem.create(com.label);
                const ted = vscode_languageserver_types_1.TextEdit.replace(com.range
                    ? com.range
                    : new completion_providers_1.ProviderRange(new completion_providers_1.ProviderPosition(position.line, Math.max(position.character - 1, 0)), position), typeof com.insertText === 'string' ? com.insertText : com.insertText.source);
                lspCompletion.insertTextFormat = 2;
                lspCompletion.detail = com.detail;
                lspCompletion.textEdit = ted;
                lspCompletion.sortText = com.sortText;
                lspCompletion.filterText =
                    typeof com.insertText === 'string' ? com.insertText : com.insertText.source;
                if (com.additionalCompletions) {
                    lspCompletion.command = vscode_languageserver_types_1.Command.create('additional', 'editor.action.triggerSuggest');
                }
                else if (com.triggerSignature) {
                    lspCompletion.command = vscode_languageserver_types_1.Command.create('additional', 'editor.action.triggerParameterHints');
                }
                return lspCompletion;
            })
                .concat(this.cssService.getCompletions(document, position));
        }
        else {
            return [];
        }
    }
    async onDefinition(params) {
        const documentUri = params.textDocument.uri;
        const docPath = vscode_uri_1.URI.parse(documentUri).fsPath;
        if (!documentUri.endsWith('.st.css') && !documentUri.startsWith('untitled:')) {
            return [];
        }
        const fileContent = this.fs.readFileSync(docPath, 'utf8');
        const pos = params.position;
        const res = await this.provider.getDefinitionLocation(fileContent, {
            line: pos.line,
            character: pos.character
        }, docPath, this.fs);
        return res.map(loc => vscode_languageserver_types_1.Location.create(vscode_uri_1.URI.file(loc.uri).toString(), loc.range));
    }
    onHover(params) {
        const doc = this.docsDispatcher.get(params.textDocument.uri);
        return doc && doc.uri.endsWith('.st.css') ? this.cssService.doHover(doc, params.position) : null;
    }
    onReferences(params) {
        const u = vscode_uri_1.URI.parse(params.textDocument.uri);
        if (!u.path.endsWith('.st.css') && !u.path.startsWith('untitled:')) {
            return [];
        }
        const refs = this.getRefs(u.fsPath, params.position);
        if (refs.length) {
            return dedupe_refs_1.dedupeRefs(refs);
        }
        else {
            const doc = this.docsDispatcher.get(params.textDocument.uri);
            return doc ? dedupe_refs_1.dedupeRefs(this.cssService.findReferences(doc, params.position)) : [];
        }
    }
    onDocumentColor(params) {
        const documentUri = params.textDocument.uri;
        if (!documentUri.endsWith('.st.css') && !documentUri.startsWith('untitled:')) {
            return [];
        }
        const doc = this.docsDispatcher.get(params.textDocument.uri);
        return doc ? color_provider_1.resolveDocumentColors(this.stylable, this.cssService, doc, this.fs) : [];
    }
    onColorPresentation(params) {
        const documentUri = params.textDocument.uri;
        if (!documentUri.endsWith('.st.css') && !documentUri.startsWith('untitled:')) {
            return [];
        }
        const doc = this.docsDispatcher.get(params.textDocument.uri);
        return doc ? color_provider_1.getColorPresentation(this.cssService, doc, params) : [];
    }
    onRenameRequest(params) {
        const documentUri = params.textDocument.uri;
        if (!documentUri.endsWith('.st.css') && !documentUri.startsWith('untitled:')) {
            return { changes: {} };
        }
        const edit = { changes: {} };
        provider_2.getRenameRefs(vscode_uri_1.URI.parse(params.textDocument.uri).fsPath, params.position, this.fs, this.stylable).forEach(ref => {
            if (edit.changes[ref.uri]) {
                edit.changes[ref.uri].push({ range: ref.range, newText: params.newName });
            }
            else {
                edit.changes[ref.uri] = [{ range: ref.range, newText: params.newName }];
            }
        });
        return edit;
    }
    onSignatureHelp(params) {
        const documentPath = vscode_uri_1.URI.parse(params.textDocument.uri).fsPath;
        if (!documentPath.endsWith('.st.css') && !documentPath.startsWith('untitled:')) {
            return null;
        }
        const doc = this.fs.readFileSync(params.textDocument.uri, 'utf8');
        const sig = this.provider.getSignatureHelp(doc, params.position, documentPath, this.fs, vscode_languageserver_types_1.ParameterInformation);
        return Promise.resolve(sig);
    }
    onDocumentFormatting() {
        // no op
        return null;
    }
}
exports.StylableLanguageService = StylableLanguageService;
async function diagnose({ connection, docsDispatcher, stylable, cssService }) {
    let res;
    let ignore = false;
    try {
        res = await connection.workspace.getConfiguration({
            section: 'stylable'
        });
        if (!!res && !!res.diagnostics && !!res.diagnostics.ignore && !!res.diagnostics.ignore.length) {
            ignore = true;
        }
    }
    catch (e) {
        /*Client has no workspace/configuration method, ignore silently */
    }
    const result = [];
    docsDispatcher.keys().forEach(key => {
        const doc = docsDispatcher.get(key);
        if (!!doc) {
            if (doc.languageId === 'stylable') {
                let diagnostics;
                if (ignore &&
                    res.diagnostics.ignore.some(p => {
                        return vscode_uri_1.URI.parse(doc.uri).fsPath.startsWith(p);
                    })) {
                    diagnostics = [];
                }
                else {
                    diagnostics = diagnosis_1.createDiagnosis(doc.getText(), vscode_uri_1.URI.parse(doc.uri).fsPath, stylable)
                        .map(diag => {
                        diag.source = 'stylable';
                        return diag;
                    })
                        .concat(cssService.getDiagnostics(doc));
                    result.push(...diagnostics);
                }
                connection.sendDiagnostics({ uri: doc.uri, diagnostics });
            }
        }
    });
    return result;
}
//# sourceMappingURL=service.js.map