"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const VCL = tslib_1.__importStar(require("vscode-css-languageservice"));
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const provider_1 = require("../lib/provider");
const vscode_uri_1 = require("vscode-uri");
function readDocRange(doc, rng) {
    const lines = doc.getText().split('\n');
    return lines[rng.start.line].slice(rng.start.character, rng.end.character);
}
function findPseudoStateStart(line, lookFrom) {
    let i = lookFrom - 1;
    let res = -1;
    let openParens = 0;
    while (i !== -1) {
        if (line[i] === ':' && line[i - 1] !== ':') {
            res = i;
        }
        if (line[i] === '(') {
            openParens++;
        }
        if (line[i] === ')') {
            openParens--;
        }
        i--;
    }
    return {
        index: res,
        openParens
    };
}
/**
 * the API for "normal" css language features fallback
 */
class CssService {
    constructor(fs) {
        this.fs = fs;
        this.inner = VCL.getCSSLanguageService();
    }
    getCompletions(document, position) {
        const cssCompsRaw = this.inner.doComplete(document, position, this.inner.parseStylesheet(document));
        return cssCompsRaw ? cssCompsRaw.items : [];
    }
    getDiagnostics(document) {
        if (!document.uri.endsWith('.css')) {
            return [];
        }
        const stylesheet = this.inner.parseStylesheet(document);
        return this.inner
            .doValidation(document, stylesheet)
            .filter(diag => {
            if (diag.code === 'emptyRules') {
                return false;
            }
            if (diag.code === 'unknownAtRules' && readDocRange(document, diag.range) === '@custom-selector') {
                return false;
            }
            if (diag.code === 'unknownAtRules' && readDocRange(document, diag.range) === '@st-scope') {
                return false;
            }
            if (diag.code === 'css-lcurlyexpected' &&
                readDocRange(document, vscode_languageserver_types_1.Range.create(vscode_languageserver_types_1.Position.create(diag.range.start.line, 0), diag.range.end)).startsWith('@custom-selector')) {
                return false;
            }
            if (diag.code === 'css-rparentexpected' || diag.code === 'css-identifierexpected') {
                const endOfLine = diag.range.end;
                endOfLine.character = -1;
                const line = readDocRange(document, vscode_languageserver_types_1.Range.create(vscode_languageserver_types_1.Position.create(diag.range.start.line, 0), endOfLine));
                const stateStart = findPseudoStateStart(line, diag.range.start.character);
                if (stateStart.index !== -1 && stateStart.openParens > 0) {
                    return false;
                }
            }
            if (diag.code === 'unknownProperties') {
                const prop = diag.message.match(/'(.*)'/)[1];
                const filePath = vscode_uri_1.URI.parse(document.uri).fsPath;
                const src = this.fs.readFileSync(filePath, 'utf8');
                const meta = provider_1.createMeta(src, filePath).meta;
                if (meta && Object.keys(meta.mappedSymbols).some(ms => ms === prop)) {
                    return false;
                }
            }
            return true;
        })
            .map(diag => {
            diag.source = 'css';
            return diag;
        });
    }
    doHover(document, position) {
        const stylesheet = this.inner.parseStylesheet(document);
        return this.inner.doHover(document, position, stylesheet);
    }
    findReferences(document, position) {
        const stylesheet = this.inner.parseStylesheet(document);
        return this.inner.findReferences(document, position, stylesheet);
    }
    getColorPresentations(document, color, range) {
        const stylesheet = this.inner.parseStylesheet(document);
        return this.inner.getColorPresentations(document, stylesheet, color, range);
    }
    findColors(document) {
        const stylesheet = this.inner.parseStylesheet(document);
        return this.inner.findDocumentColors(document, stylesheet);
    }
    findColor(document) {
        const colors = this.findColors(document);
        return colors.length ? colors[0].color : null;
    }
}
exports.CssService = CssService;
//# sourceMappingURL=css-service.js.map